@@ -7,6 +7,8 @@ use cached::proc_macro::cached;
 use derive_builder::Builder;
 use qcs_api_client_common::configuration::RefreshError;
 pub use qcs_api_client_grpc::channel::Error as GrpcError;
+#[cfg(feature = "grpc-web")]
+use qcs_api_client_grpc::channel::wrap_channel_with_grpc_web;
 use qcs_api_client_grpc::{
     channel::{parse_uri, wrap_channel_with, wrap_channel_with_retry},
     get_channel_with_timeout,
@@ -77,12 +79,15 @@ impl From<String> for JobId {
 
 /// Execute compiled program on a QPU.
 ///
+/// See [`ExecuteControllerJobRequest`] for more details.
+///
 /// # Arguments
 /// * `quantum_processor_id` - The quantum processor to execute the job on. This parameter
 ///      is required unless using [`ConnectionStrategy::EndpointId`] in `execution_options`
 ///      to target a specific endpoint ID.
 /// * `program` - The compiled program as an [`EncryptedControllerJob`]
-/// * `patch_values` - The parameters to use for the execution.
+/// * `patch_values` - The parameters to use for the execution. See [`submit_with_parameter_batch`]
+///      if you need to execute with multiple sets of parameters.
 /// * `client` - The [`Qcs`] client to use.
 /// * `execution_options` - The [`ExecutionOptions`] to use. If the connection strategy used
 ///       is [`ConnectionStrategy::EndpointId`] then direct access to that endpoint
@@ -94,6 +99,50 @@ pub async fn submit(
     client: &Qcs,
     execution_options: &ExecutionOptions,
 ) -> Result<JobId, QpuApiError> {
+    submit_with_parameter_batch(
+        quantum_processor_id,
+        program,
+        std::iter::once(patch_values),
+        client,
+        execution_options,
+    )
+    .await?
+    .pop()
+    .ok_or_else(|| GrpcClientError::ResponseEmpty("Job Execution ID".into()))
+    .map_err(QpuApiError::from)
+}
+
+/// Execute a compiled program on a QPU with multiple sets of `patch_values`.
+///
+/// See [`ExecuteControllerJobRequest`] for more details.
+///
+/// # Arguments
+/// * `quantum_processor_id` - The quantum processor to execute the job on. This parameter
+///      is required unless using [`ConnectionStrategy::EndpointId`] in `execution_options`
+///      to target a specific endpoint ID.
+/// * `program` - The compiled program as an [`EncryptedControllerJob`]
+/// * `patch_values` - The parameters to use for the execution. The job will be run once for each
+///     given set of [`Parameters`].
+/// * `client` - The [`Qcs`] client to use.
+/// * `execution_options` - The [`ExecutionOptions`] to use. If the connection strategy used
+///       is [`ConnectionStrategy::EndpointId`] then direct access to that endpoint
+///       overrides the `quantum_processor_id` parameter.
+///
+/// # Errors
+///
+/// * Returns a [`QpuApiError`] if:
+///     * Any of the jobs fail to be queued.
+///     * The provided `patch_values` iterator is empty.
+pub async fn submit_with_parameter_batch<'a, I>(
+    quantum_processor_id: Option<&str>,
+    program: EncryptedControllerJob,
+    patch_values: I,
+    client: &Qcs,
+    execution_options: &ExecutionOptions,
+) -> Result<Vec<JobId>, QpuApiError>
+where
+    I: IntoIterator<Item = &'a Parameters>,
+{
     #[cfg(feature = "tracing")]
     tracing::debug!(
         "submitting job to {:?} using options {:?}",
@@ -101,8 +150,15 @@ pub async fn submit(
         execution_options
     );
 
+    let mut patch_values = patch_values.into_iter().peekable();
+    if patch_values.peek().is_none() {
+        return Err(QpuApiError::EmptyPatchValues);
+    }
+
     let request = ExecuteControllerJobRequest {
-        execution_configurations: vec![params_into_job_execution_configuration(patch_values)],
+        execution_configurations: patch_values
+            .map(params_into_job_execution_configuration)
+            .collect(),
         job: Some(execute_controller_job_request::Job::Encrypted(program)),
         target: execution_options.get_job_target(quantum_processor_id),
         options: execution_options.api_options().cloned(),
@@ -112,18 +168,15 @@ pub async fn submit(
         .get_controller_client(client, quantum_processor_id)
         .await?;
 
-    // we expect exactly one job ID since we only submit one execution configuration
-    let job_execution_id = controller_client
+    Ok(controller_client
         .execute_controller_job(request)
         .await
         .map_err(GrpcClientError::RequestFailed)?
         .into_inner()
         .job_execution_ids
-        .pop();
-
-    Ok(job_execution_id
+        .into_iter()
         .map(JobId)
-        .ok_or_else(|| GrpcClientError::ResponseEmpty("Job Execution ID".into()))?)
+        .collect())
 }
 
 /// Cancel all given jobs that have yet to begin executing.
@@ -478,10 +531,11 @@ impl ExecutionOptions {
         let uri = parse_uri(address).map_err(QpuApiError::GrpcError)?;
         let channel = get_channel_with_timeout(uri, self.timeout())
             .map_err(|err| QpuApiError::GrpcError(err.into()))?;
-        Ok(wrap_channel_with_retry(wrap_channel_with(
-            channel,
-            client.get_config().clone(),
-        )))
+        let channel =
+            wrap_channel_with_retry(wrap_channel_with(channel, client.get_config().clone()));
+        #[cfg(feature = "grpc-web")]
+        let channel = wrap_channel_with_grpc_web(channel);
+        Ok(channel)
     }
 
     async fn get_gateway_address(
@@ -622,6 +676,10 @@ pub enum QpuApiError {
     #[error("A quantum processor ID must be provided if not connecting directly to an endpoint ID with ConnectionStrategy::EndpointId")]
     MissingQpuId,
 
+    /// Error due to user not providing patch values
+    #[error("Submitting a job requires at least one set of patch values")]
+    EmptyPatchValues,
+
     /// Error that can occur when controller service fails to execute a job
     #[error("The submitted job failed with status: {status}. {message}")]
     JobExecutionFailed {
